# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


'''
Simple sequential processor mapping for dispel4py graphs.
This processor determines the dependencies of each PE in the graph and
executes them sequentially.

From the commandline, run the following command::

    dispel4py simple <module|module file> \\
              [-a attribute] \\
              [-f inputfile] \\
              [-d inputdata] \\
              [-i iterations] \\
              [-h]

with parameters

:module: module that creates a Dispel4Py graph
:-a attr:   name of the graph attribute within the module (optional)
:-f file:   file containing input data in JSON format (optional)
:-d data:   input data in JSON format (optional)
:-i iter:   number of iterations to compute (default is 1)
:-h:        print this help page

The input data must be a dictionary mapping either a PE name or an PE
identifier to a list of input data or the number of iterations which is a
non-negative integer.

For example::

    dispel4py simple dispel4py.examples.graph_testing.pipeline_test -i 5

    Processing 5 iterations.
    Inputs: {'TestProducer0': 5}
    SimplePE: Processed 1 iteration.
    Outputs: {'TestOneInOneOut5': {'output': [1, 2, 3, 4, 5]}}

The simple processor can also be called from an interactive Python session.
For example, the following source creates a simple graph with a producer and a
consumer PE and executes 5 iterations::

    from dispel4py.workflow_graph import WorkflowGraph
    from dispel4py.examples.graph_testing.testing_PEs \\
        import TestProducer, TestOneInOneOut

    graph = WorkflowGraph()
    prod = TestProducer()
    cons = TestOneInOneOut()
    graph.connect(prod, 'output', cons, 'input')

    from dispel4py.new.simple_process import process_and_return
    results = process_and_return(graph, { prod : 5 })

The return value is the collection the output data of the consumer PE as this
output stream is not connected::

    print results
    {'TestOneInOneOut1': {'output': [1, 2, 3, 4, 5]}}
'''

import types
from dispel4py.new.processor import GenericWrapper, SimpleProcessingPE
from dispel4py.new import processor


def simpleLogger(self, msg):
    print("%s: %s" % (self.id, msg))


def process_and_return(workflow, inputs, resultmappings=None):
    '''
    Executes the simple sequential processor for dispel4py graphs and returns
    the data collected from any unconnected output streams.

    :param workflow: the dispel4py graph to be enacted
    :param inputs: inputs for root PEs of the graphs.
        This is a dictionary mapping a PE to either a non-negative integer
        (the number of iterations) or a list of input data items.
    :rtype: a dictionary mapping PE ids to the output data produced by that PE

    '''
    numnodes = 0
    for node in workflow.graph.nodes():
        numnodes += 1
        node.getContainedObject().numprocesses = 1
    processes, inputmappings, outputmappings = \
        processor.assign_and_connect(workflow, numnodes)
    # print 'Processes: %s' % processes
    # print inputmappings
    # print outputmappings
    proc_to_pe = {}
    for node in workflow.graph.nodes():
        pe = node.getContainedObject()
        proc_to_pe[processes[pe.id][0]] = pe

    simple = SimpleProcessingPE(inputmappings, outputmappings, proc_to_pe)
    simple.id = 'SimplePE'
    simple.result_mappings = resultmappings
    wrapper = SimpleProcessingWrapper(simple, [inputs])
    wrapper.targets = {}
    wrapper.sources = {}
    wrapper.process()

    # now collect output data into a single list for each PE
    outputs = {}
    for (pe_id, output_name), data in wrapper.outputs.items():
        if pe_id not in outputs:
            outputs[pe_id] = {}
        try:
            outputs[pe_id][output_name] += data
        except KeyError:
            outputs[pe_id][output_name] = data
    return outputs


def process(workflow, inputs, args=None, resultmappings=None):
    '''
    Executes the simple sequential processor for dispel4py graphs and prints
    out the input and output data. This is the default target when invoking
    the simple mapping with `dispel4py simple <module>`.

    :param workflow: the dispel4py graph to be enacted
    :param inputs: inputs for root PEs of the graphs. This is a dictionary
        mapping a PE id to either a non-negative integer
        (the number of iterations) or a list of input data items.
    '''

    try:
        print('Inputs: %s' % {pe.id: data for pe, data in inputs.items()})
    except:
        print('Inputs: %s' % {pe: data for pe, data in inputs.items()})
    results = process_and_return(workflow, inputs, resultmappings)
    print('Outputs: %s' % results)


class SimpleProcessingWrapper(GenericWrapper):

    def __init__(self, pe, provided_inputs=None):
        GenericWrapper.__init__(self, pe)
        self.pe.log = types.MethodType(simpleLogger, pe)
        self.provided_inputs = provided_inputs
        self.outputs = {}

    def _read(self):
        result = super(SimpleProcessingWrapper, self)._read()
        return result

    def _write(self, name, data):
        # self.pe.log('Writing %s to %s' % (data, name))
        try:
            self.outputs[name].extend(data)
        except KeyError:
            self.outputs[name] = data


def main():                                                  # pragma: no cover
    '''
    This is the commandline entry point for the simple processor.
    '''
    from dispel4py.new.processor \
        import load_graph_and_inputs, create_arg_parser

    parser = create_arg_parser()
    args = parser.parse_args()

    graph, inputs = load_graph_and_inputs(args)
    if graph is not None:
        process(graph, inputs)



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dispel4py.base import IterativePE, ConsumerPE, CompositePE

from collections import defaultdict


class SplitTextFile(IterativePE):

    def __init__(self):
        IterativePE.__init__(self)

    def _process(self, filename):
        with open(filename) as f:
            for line in f:
                self.write('output', line.strip())


class WordCount(IterativePE):

    def __init__(self):
        IterativePE.__init__(self)

    def _process(self, line):
        for w in line.split(' '):
            self.write('output', (w, 1))


class CountByGroup(IterativePE):

    def __init__(self):
        IterativePE.__init__(self)
        self.count = defaultdict(int)

    def _process(self, data):
        self.count[data[0]] += data[1]

    def _postprocess(self):
        for (word, count) in self.count.items():
            self.write('output', (word, count))


class Print(ConsumerPE):

    def __init__(self):
        ConsumerPE.__init__(self)

    def _process(self, data):
        self.log('%s: %s' % data)


def count_by_group():
    composite = CompositePE()
    count = CountByGroup()
    merge = CountByGroup()
    merge._add_input('input', grouping='global')
    composite.connect(count, 'output', merge, 'input')
    composite._map_input('input', count, 'input')
    composite._map_output('output', merge, 'output')
    return composite


from dispel4py.workflow_graph import WorkflowGraph

graph = WorkflowGraph()
textfile = SplitTextFile()
wordcount = WordCount()
# count = CountByGroup()
# merge = CountByGroup()
# merge._add_input('input', grouping='global')
count = count_by_group()
result = Print()

graph.connect(textfile, 'output', wordcount, 'input')
graph.connect(wordcount, 'output', count, 'input')
# graph.connect(count, 'output', merge, 'input')
# graph.connect(merge, 'output', result, 'input')
graph.connect(count, 'output', result, 'input')



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Test PEs
'''
Example PEs for test workflows, implementing various patterns.
'''

from dispel4py.core import GenericPE
from dispel4py.base import IterativePE, ProducerPE, ConsumerPE
import random
import time
from collections import defaultdict


class TestProducer(GenericPE):
    '''
    This PE produces a range of numbers
    '''

    def __init__(self, numOutputs=1):
        GenericPE.__init__(self)
        if numOutputs == 1:
            self._add_output('output', tuple_type=['number'])
        else:
            for i in range(numOutputs):
                self._add_output('output%s' % i, tuple_type=['number'])
        self.counter = 0
        self.outputnames = list(self.outputconnections.keys())

    def _process(self, inputs):
        self.counter += 1
        result = {}
        for output in self.outputnames:
            result[output] = self.counter
        # self.log("Writing out %s" % result)
        return result


class NumberProducer(GenericPE):

    def __init__(self, numIterations=1):
        GenericPE.__init__(self)
        self._add_output('output', tuple_type=['number'])
        self.counter = 0
        self.numIterations = numIterations

    def _process(self, inputs):
        for i in range(self.numIterations):
            self.write('output', [self.counter*i+i])
        self.counter += 1


class IntegerProducer(ProducerPE):

    def __init__(self, start, limit):
        ProducerPE.__init__(self)
        self.start = start
        self.limit = limit

    def _process(self, inputs):
        for i in range(self.start, self.limit):
            self.write('output', i)


class TestOneInOneOut(GenericPE):
    '''
    This PE outputs the input data.
    '''

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input')
        self._add_output('output')

    def setInputTypes(self, inputtypes):
        self.inputconnections['input']['type'] = inputtypes['input']
        self.outputconnections['output']['type'] = inputtypes['input']

    def process(self, inputs):
        # self.log('Processing inputs %s' % inputs)
        return {'output': inputs['input']}


class TestIterative(IterativePE):
    '''
    This PE outputs the input data.
    '''

    def __init__(self):
        IterativePE.__init__(self)

    def _process(self, data):
        return data


class TestDelayOneInOneOut(GenericPE):
    '''
    This PE outputs the input data.
    '''

    def __init__(self, delay=1):
        GenericPE.__init__(self)
        self._add_input('input')
        self._add_output('output', tuple_type=['number'])
        self.delay = delay

    def process(self, inputs):
        # self.log('Processing inputs %s' % inputs)
        time.sleep(self.delay)
        return {'output': inputs['input']}


class TestOneInOneOutWriter(GenericPE):
    '''
    This PE copies the input to an output, but it uses the write method.
    Remember that the write function allows to produce more than one output
    block within one processing step.
    '''

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input')
        self._add_output('output', tuple_type=['number'])

    def process(self, inputs):
        self.write('output', inputs['input'])


class TestTwoInOneOut(GenericPE):
    '''
    This PE takes two inputs and it merges the data into one output string.
    '''

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input0')
        self._add_input('input1')
        self._add_output('output', tuple_type=['result'])

    def process(self, inputs):
        # print '%s: inputs %s' % (self.id, inputs)
        result = ''
        for inp in self.inputconnections:
            if inp in inputs:
                result += '%s' % (inputs[inp])
        if result:
            # print '%s: result %s' % (self.id, result)
            return {'output': result}


class TestMultiProducer(GenericPE):

    def __init__(self, num_output=10):
        GenericPE.__init__(self)
        self._add_output('output')
        self.num_output = num_output

    def _process(self, inputs):
        for i in range(self.num_output):
            self.write('output', i)


class PrintDataConsumer(ConsumerPE):

    def __init__(self):
        ConsumerPE.__init__(self)

    def _process(self, data):
        print(data)


class RandomFilter(GenericPE):
    '''
    This PE randomly filters the input.
    '''
    input_name = 'input'
    output_name = 'output'

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input')
        self._add_output('output', tuple_type=['word'])

    def process(self, inputs):
        if random.choice([True, False]):
            return {'output': inputs['input']}
            # self.write('output', inputs['input'] )
        return None


class WordCounter(GenericPE):
    '''
    This PE counts the number of times (counter) that it receives each word.
    And it produces as an output: the same word (the input) and its counter.
    '''
    input_name = 'input'
    output_name = 'output'

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input', grouping=[0])
        self._add_output('output', tuple_type=['word', 'count'])
        self.mywords = defaultdict(int)

    def _process(self, inputs):
        word = inputs['input'][0]
        self.mywords[word] += 1
        return {'output': [word, self.mywords[word]]}


class RandomWordProducer(GenericPE):
    '''
    This PE produces a random word as an output.
    '''
    words = ["dispel4py", "computing", "mpi", "processing",
             "simple", "analysis", "data"]

    def __init__(self):
        GenericPE.__init__(self)
        self._add_output('output', tuple_type=['word'])

    def process(self, inputs=None):
        word = random.choice(RandomWordProducer.words)
        outputs = {'output': [word]}
        return outputs



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This graph is a modification of the
:py:mod:`~test.graph_testing.parallel_pipeline` example,
showing how the user can specify how the graph is going to be partitioned
into MPI processes.
In this example we are specifying that one MPI process is executing the
pipeline of nodes ``prod``, ``cons1`` and ``cons2`` and the other MPI
processes are executing the remaining node ``cons3``::

    graph.partitions = [ [prod, cons1, cons2], [cons3] ]


It can be executed with MPI and Storm. Storm will ignore the partition
information.

Execution:

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process,
    and the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 3 python -m dispel4py.worker_mpi \\
            dispel4py.examples.graph_testing.partition_parallel_pipeline -i 10

    Output::

        Partitions:  [TestProducer0, TestOneInOneOut1, TestOneInOneOut2], \
[TestOneInOneOut3]
        Processes: {'GraphWrapperPE5': [0, 1], 'GraphWrapperPE4': [2]}
        GraphWrapperPE5 (rank 0): I'm a bolt
        GraphWrapperPE5 (rank 1): I'm a bolt
        GraphWrapperPE4 (rank 2): I'm a spout
        Rank 2: Sending terminate message to [0, 1]
        GraphWrapperPE4 (rank 2): Processed 10 input block(s)
        GraphWrapperPE4 (rank 2): Completed.
        GraphWrapperPE5 (rank 1): Processed 5 input block(s)
        GraphWrapperPE5 (rank 1): Completed.
        GraphWrapperPE5 (rank 0): Processed 5 input block(s)
        GraphWrapperPE5 (rank 0): Completed.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testParallelPipeline():
    '''
    Creates the parallel pipeline graph with partitioning information.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    prod = t.TestProducer()
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOut()
    cons3 = t.TestOneInOneOut()

    graph.connect(prod, 'output', cons1, 'input')
    graph.connect(cons1, 'output', cons2, 'input')
    graph.connect(cons1, 'output', cons3, 'input')

    graph.partitions = [[prod, cons1, cons2], [cons3]]

    return graph

''' important: this is the graph_variable '''
graph = testParallelPipeline()



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a workflow which copies the data
(from node ``prod``) to one node (``cons``).

Note, that in this graph, we have decided to assign two processes to instances
of the same PE::

    cons = t.TestOneInOneOut()
    cons.numprocesses=2

Another interesting point in this example is how to define the different types
of groupings. In this example we have::

    cons.inputconnections['input']['grouping'] = 'all'

which means that the ``prod`` node sends copies of its output data to *all* the
connected instances.
For that reason, the node ``cons`` has to specify the input grouping 'all' to
the connection.
In :py:mod:`~test.graph_testing.grouping_split_merge`, we use another type of
grouping, which is group by. However, in that case, the grouping type is
defined by the PE class (WordCounter) which means it applies to all instances
of that class (unless it is explicitly overridden by an instance as we did
above).

If you compare this graph with :py:mod:`~test.graph_testing.teecopy` these
look quite similar. However, they do different things.
In this example, we have two instances of the same PE and the node ``prod``
sends the same data to both instances whereas in
:py:mod:`~test.graph_testing.teecopy` the node ``prod`` sends the same data
to two different PEs.

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> python -m dispel4py.worker_mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When <-i number of interations/runs> is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 3 python -m dispel4py.worker_mpi\\
            dispel4py.examples.graph_testing.grouping_onetoall

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 3 nodes. For this reason we need at least 3 MPI
        processes to execute it.

    Output::

        Processing 1 iterations
        Processes: {'TestProducer0': [2], 'TestOneInOneOut1': [0, 1]}
        TestOneInOneOut1 (rank 0): I'm a bolt
        TestOneInOneOut1 (rank 1): I'm a bolt
        TestProducer0 (rank 2): I'm a spout
        Rank 2: Sending terminate message to [0, 1]
        TestProducer0 (rank 2): Processed 1 input block(s)
        TestProducer0 (rank 2): Completed.
        TestOneInOneOut1 (rank 1): Processed 1 input block(s)
        TestOneInOneOut1 (rank 1): Completed.
        TestOneInOneOut1 (rank 0): Processed 1 input block(s)
        TestOneInOneOut1 (rank 0): Completed.


* STORM:
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testOnetoAll():
    graph = WorkflowGraph()
    prod = t.TestProducer()
    cons = t.TestOneInOneOut()
    cons.numprocesses = 2
    cons.inputconnections['input']['grouping'] = 'all'
    graph.connect(prod, 'output', cons, 'input')
    return graph


''' important: this is the graph_variable '''
graph = testOnetoAll()



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Enactment of dispel4py graphs.
This module contains methods that are used by different mappings.

From the commandline, run the following command::

    dispel4py <mapping> <module>  [-h] [-a attribute] [-f inputfile]\
                                  [-i iterations] [...]

with parameters

:mapping:   target mapping
:module:    module that creates a Dispel4Py graph
:-n num:    number of processes (required)
:-a attr:   name of the graph attribute within the module (optional)
:-f file:   file containing input data in JSON format (optional)
:-i iter:   number of iterations to compute (default is 1)
:-h:        print this help page

Other parameters might be required by the target mapping, for example the
number of processes if running in a parallel environment.

'''

import argparse
import os
import types

from dispel4py.core import GROUPING
from dispel4py.utils import make_hash
from dispel4py.new.mappings import config

STATUS_ACTIVE = 10
STATUS_INACTIVE = 11
STATUS_TERMINATED = 12
# mapping for name to value
STATUS = {STATUS_ACTIVE: 'ACTIVE',
          STATUS_INACTIVE: 'INACTIVE',
          STATUS_TERMINATED: 'TERMINATED'}


def simpleLogger(self, msg):
    try:
        print("%s (rank %s): %s" % (self.id, self.rank, msg))
    except:
        print("%s: %s" % (self.id, msg))


def get_inputs(pe, inputs):
    try:
        return inputs[pe]
    except KeyError:
        pass
    try:
        return inputs[pe.name]
    except:
        pass
    try:
        return inputs[pe.id]
    except:
        pass


class GenericWriter(object):
    def __init__(self, wrapper, name):
        self.wrapper = wrapper
        self.name = name

    def write(self, data):
        self.wrapper._write(self.name, data)


class GenericWrapper(object):

    def __init__(self, pe):
        self.pe = pe
        self.pe.wrapper = self
        for o in self.pe.outputconnections:
            self.pe.outputconnections[o]['writer'] = GenericWriter(self, o)
        self.targets = {}
        self._sources = {}

    @property
    def sources(self):
        return self._sources

    @sources.setter
    def sources(self, sources):
        # count and store number of inputs when setting the sources
        num_inputs = 0
        for i in sources.values():
            num_inputs += len(i)
        self._num_sources = num_inputs
        self._sources = sources

    def process(self):
        num_iterations = 0
        self.pe.preprocess()
        result = self._read()
        inputs, status = result
        # self.pe.log('Result: %s, status=%s' % (inputs, STATUS[status]))
        while status != STATUS_TERMINATED:
            if inputs is not None:
                outputs = self.pe.process(inputs)
                num_iterations += 1
                if outputs is not None:
                    # self.pe.log('Produced output: %s' % outputs)
                    for key, value in outputs.items():
                        self._write(key, value)
            inputs, status = self._read()
            # self.pe.log('Result: %s, status=%s' % (inputs, STATUS[status]))
        self.pe.postprocess()
        self._terminate()
        if num_iterations == 1:
            self.pe.log('Processed 1 iteration.')
        else:
            self.pe.log('Processed %s iterations.' % num_iterations)

    def _read(self):
        # check the provided inputs
        if self.provided_inputs is not None:
            if isinstance(self.provided_inputs, int) and \
                    self.provided_inputs > 0:
                self.provided_inputs -= 1
                return {}, STATUS_ACTIVE
            elif self.provided_inputs:
                return self.provided_inputs.pop(0), STATUS_ACTIVE
            else:
                return None, STATUS_TERMINATED

    def _write(self, name, data):
        None

    def _terminate(self):
        None


class ShuffleCommunication(object):
    def __init__(self, rank, sources, destinations):
        self.destinations = destinations
        self.currentIndex = (sources.index(rank) % len(self.destinations)) - 1
        self.name = None

    def getDestination(self, data):
        self.currentIndex = (self.currentIndex + 1) % len(self.destinations)
        return [self.destinations[self.currentIndex]]


class GroupByCommunication(object):
    def __init__(self, destinations, input_name, groupby):
        self.groupby = groupby
        self.destinations = destinations
        self.input_name = input_name
        self.name = groupby

    def getDestination(self, data):
        output = tuple([data[self.input_name][x] for x in self.groupby])
        dest_index = abs(make_hash(output)) % len(self.destinations)
        return [self.destinations[dest_index]]


class AllToOneCommunication(object):
    def __init__(self, destinations):
        self.destinations = destinations
        self.name = 'global'

    def getDestination(self, data):
        return [self.destinations[0]]


class OneToAllCommunication(object):
    def __init__(self, destinations):
        self.destinations = destinations
        self.name = 'all'

    def getDestination(self, data):
        return self.destinations


def _getConnectedInputs(node, graph):
    names = []
    for edge in graph.edges(node, data=True):
        direction = edge[2]['DIRECTION']
        dest = direction[1]
        dest_input = edge[2]['TO_CONNECTION']
        if dest == node.getContainedObject():
            names.append(dest_input)
    return names


def _getNumProcesses(size, numSources, numProcesses, totalProcesses):
    div = max(1, totalProcesses - numSources)
    return int(numProcesses * (size - numSources) / div)


def _assign_processes(workflow, size):
    graph = workflow.graph
    processes = {}
    success = True
    totalProcesses = 0
    numSources = 0
    sources = []
    for node in graph.nodes():
        pe = node.getContainedObject()
        # if pe.inputconnections:
        if _getConnectedInputs(node, graph):
            totalProcesses = totalProcesses + pe.numprocesses
        else:
            sources.append(pe.id)
            totalProcesses += 1
            numSources += 1

    if totalProcesses > size:
        success = False
        # we need at least one process for each node in the graph
        print('Graph is larger than job size: %s > %s.' %
              (totalProcesses, size))
    else:
        node_counter = 0
        for node in graph.nodes():
            pe = node.getContainedObject()
            prcs = 1 if pe.id in sources else _getNumProcesses(
                size, numSources, pe.numprocesses, totalProcesses)
            processes[pe.id] = range(node_counter, node_counter + prcs)
            node_counter = node_counter + prcs
    return success, sources, processes


def _getCommunication(rank, source_processes,
                      dest, dest_input, dest_processes):
    communication = ShuffleCommunication(
        rank, source_processes, dest_processes)
    try:
        if GROUPING in dest.inputconnections[dest_input]:
            groupingtype = dest.inputconnections[dest_input][GROUPING]
            if isinstance(groupingtype, list):
                communication = GroupByCommunication(
                    dest_processes, dest_input, groupingtype)
            elif groupingtype == 'all':
                communication = OneToAllCommunication(dest_processes)
            elif groupingtype == 'global':
                communication = AllToOneCommunication(dest_processes)
    except KeyError:
        print("No input '%s' defined for PE '%s'" % (dest_input, dest.id))
        raise
    return communication


def _create_connections(graph, node, processes):
    pe = node.getContainedObject()
    inputmappings = {i: {} for i in processes[pe.id]}
    outputmappings = {i: {} for i in processes[pe.id]}
    for edge in graph.edges(node, data=True):
        direction = edge[2]['DIRECTION']
        source = direction[0]
        source_output = edge[2]['FROM_CONNECTION']
        dest = direction[1]
        dest_processes = list(processes[dest.id])
        source_processes = list(processes[source.id])
        dest_input = edge[2]['TO_CONNECTION']
        allconnections = edge[2]['ALL_CONNECTIONS']
        if dest == pe:
            for i in processes[pe.id]:
                for (source_output, dest_input) in allconnections:
                    try:
                        inputmappings[i][dest_input] += source_processes
                    except KeyError:
                        inputmappings[i][dest_input] = source_processes
        if source == pe:
            for i in processes[pe.id]:
                for (source_output, dest_input) in allconnections:
                    communication = _getCommunication(
                        i, source_processes, dest, dest_input, dest_processes)
                    try:
                        outputmappings[i][source_output].append(
                            (dest_input, communication))
                    except KeyError:
                        outputmappings[i][source_output] = \
                            [(dest_input, communication)]
    return inputmappings, outputmappings


def _connect(workflow, processes):
    graph = workflow.graph
    outputmappings = {}
    inputmappings = {}
    for node in graph.nodes():
        inc, outc = _create_connections(graph, node, processes)
        inputmappings.update(inc)
        outputmappings.update(outc)
    return inputmappings, outputmappings


def assign_and_connect(workflow, size):
    success, sources, processes = _assign_processes(workflow, size)
    if success:
        inputmappings, outputmappings = _connect(workflow, processes)
        return processes, inputmappings, outputmappings
    else:
        return None

import copy

from dispel4py.workflow_graph import WorkflowGraph


def get_partitions(workflow):
    try:
        partitions = workflow.partitions
    except AttributeError:
        sourcePartition = []
        otherPartition = []
        graph = workflow.graph
        for node in graph.nodes():
            pe = node.getContainedObject()
            if not _getConnectedInputs(node, graph):
                sourcePartition.append(pe)
            else:
                otherPartition.append(pe)
        partitions = [sourcePartition, otherPartition]
        workflow.partitions = partitions
    return partitions


def create_partitioned(workflow_all):
    processes_all, inputmappings_all, outputmappings_all = \
        assign_and_connect(workflow_all, len(workflow_all.graph.nodes()))
    proc_to_pe_all = {v[0]: k for k, v in processes_all.items()}
    partitions = get_partitions(workflow_all)
    external_connections = []
    pe_to_partition = {}
    partition_pes = []
    for i in range(len(partitions)):
        for pe in partitions[i]:
            pe_to_partition[pe.id] = i
    for index in range(len(partitions)):
        result_mappings = {}
        part = partitions[index]
        partition_id = index
        component_ids = [pe.id for pe in part]
        workflow = copy.deepcopy(workflow_all)
        graph = workflow.graph
        for node in graph.nodes():
            if node.getContainedObject().id not in component_ids:
                graph.remove_node(node)
        processes, inputmappings, outputmappings = \
            assign_and_connect(workflow, len(graph.nodes()))
        proc_to_pe = {}
        for node in graph.nodes():
            pe = node.getContainedObject()
            proc_to_pe[processes[pe.id][0]] = pe
        for node in graph.nodes():
            pe = node.getContainedObject()
            pe.rank = index
            proc_all = processes_all[pe.id][0]
            for output_name in outputmappings_all[proc_all]:
                for dest_input, comm_all in\
                        outputmappings_all[proc_all][output_name]:
                    dest = proc_to_pe_all[comm_all.destinations[0]]
                    if dest not in processes:
                        # it's an external connection
                        external_connections.append((comm_all,
                                                     partition_id,
                                                     pe.id,
                                                     output_name,
                                                     pe_to_partition[dest],
                                                     dest, dest_input))
                        try:
                            result_mappings[pe.id].append(output_name)
                        except:
                            result_mappings[pe.id] = [output_name]
        partition_pe = SimpleProcessingPE(inputmappings,
                                          outputmappings,
                                          proc_to_pe)

        # use number of processes if specified in graph
        try:
            partition_pe.numprocesses = workflow_all.numprocesses[partition_id]
        except:
            # use default assignment of processes
            pass

        partition_pe.workflow = workflow
        partition_pe.partition_id = partition_id
        if result_mappings:
            partition_pe.result_mappings = result_mappings
        partition_pe.map_inputs = _map_inputs_to_pes
        partition_pe.map_outputs = _map_outputs_from_pes
        partition_pes.append(partition_pe)
    # print 'EXTERNAL CONNECTIONS : %s' % external_connections
    ubergraph = WorkflowGraph()
    ubergraph.pe_to_partition = pe_to_partition
    ubergraph.partition_pes = partition_pes
    # sort the external connections so that nodes are added in the same order
    # if doing this in multiple processes in parallel this is important
    for comm, source_partition, source_id, source_output, dest_partition, \
        dest_id, dest_input in \
            sorted(external_connections,
                   key=lambda connections: len(connections)):
        partition_pes[source_partition]._add_output((source_id, source_output))
        partition_pes[dest_partition]._add_input((dest_id, dest_input),
                                                 grouping=comm.name)
        ubergraph.connect(partition_pes[source_partition],
                          (source_id, source_output),
                          partition_pes[dest_partition],
                          (dest_id, dest_input))
    return ubergraph


def map_inputs_to_partitions(ubergraph, inputs):
    mapped_input = {}
    for pe in inputs:
        try:
            partition_id = ubergraph.pe_to_partition[pe]
            pe_id = pe
        except:
            try:
                partition_id = ubergraph.pe_to_partition[pe.id]
                pe_id = pe.id
            except Exception as exc:
                raise Exception('Could not map input name "%s" to a PE. %s: %s'
                                % (pe, exc.__class__.__name__, exc))
        mapped_pe = ubergraph.partition_pes[partition_id]
        try:
            mapped_pe_input = []
            for i in inputs[pe]:
                mapped_data = {(pe_id, name):
                               [data] for name, data in i.items()}
                mapped_pe_input.append(mapped_data)
        except TypeError:
            mapped_pe_input = inputs[pe]
        mapped_input[mapped_pe] = mapped_pe_input
    return mapped_input


def _map_inputs_to_pes(data):
    result = {}
    for i in data:
        pe_id, input_name = i
        mapped_data = [{input_name: block} for block in data[i]]
        try:
            result[pe_id].update(mapped_data)
        except KeyError:
            result[pe_id] = mapped_data
    return result


def _map_outputs_from_pes(data):
    result = {}
    for pe_id in data:
        for i in data[pe_id]:
            result[(pe_id, i)] = data[pe_id][i]
    return result


def _no_map(data):
    return data


from dispel4py.core import GenericPE


def _is_root(node, workflow):
    result = True
    pe = node.getContainedObject()
    for edge in workflow.graph[node].values():
        if pe == edge['DIRECTION'][1]:
            result = False
            break
    return result


def _get_dependencies(proc, inputmappings):
    dep = []
    for input_name, sources in inputmappings[proc].items():
        for s in sources:
            sdep = _get_dependencies(s, inputmappings)
            for n in sdep:
                if n not in dep:
                    dep.append(n)
            if s not in dep:
                dep.append(s)
    return dep


def _order_by_dependency(inputmappings, outputmappings):
    ordered = []
    for proc in outputmappings:
        if not outputmappings[proc]:
            dep = _get_dependencies(proc, inputmappings)
            for n in ordered:
                try:
                    dep.remove(n)
                except:
                    # never mind if the element wasn't in the list
                    pass
            ordered += dep
            ordered.append(proc)
    return ordered


class SimpleProcessingPE(GenericPE):
    '''
    A PE that processes a subgraph of PEs in sequence.
    '''
    def __init__(self, input_mappings, output_mappings, proc_to_pe):
        GenericPE.__init__(self)
        # work out the order of PEs
        self.ordered = _order_by_dependency(input_mappings, output_mappings)
        self.input_mappings = input_mappings
        self.output_mappings = output_mappings
        self.proc_to_pe = proc_to_pe
        self.result_mappings = None
        self.map_inputs = _no_map
        self.map_outputs = _no_map

    def _preprocess(self):
        for proc in self.ordered:
            pe = self.proc_to_pe[proc]
            try:
                pe.rank = self.rank
            except:
                pass
            pe.log = types.MethodType(simpleLogger, pe)
            pe.preprocess()

    def _postprocess(self):
        all_inputs = {}
        results = {}
        for proc in self.ordered:
            pe = self.proc_to_pe[proc]
            pe.writer = SimpleWriter(self, pe,
                                     self.output_mappings[proc],
                                     self.result_mappings)
            pe._write = types.MethodType(_simple_write, pe)
            # if there was data produced in postprocessing
            # then we need to process that data in the PEs downstream
            if proc in all_inputs:
                for data in all_inputs[proc]:
                    # pe.log('Processing input: %s' % data)
                    result = pe.process(data)
                    # pe.log('Produced result: %s' % result)
                    if result is not None:
                        for output_name in result:
                            pe.write(output_name, result[output_name])
            # once all the input data is processed this PE can finish
            pe.postprocess()
            # PE might write data during postprocessing
            for p, input_data in pe.writer.all_inputs.items():
                try:
                    all_inputs[p].extend(input_data)
                except:
                    all_inputs[p] = input_data
            if pe.writer.results:
                results[pe] = pe.writer.results
            pe.writer.all_inputs = {}
            pe.writer.results = {}
        results = self.map_outputs(results)
        for key, value in results.items():
            self._write(key, value)

    def _process(self, inputs):
        all_inputs = {}
        results = {}
        inputs = self.map_inputs(inputs)
        for proc in self.ordered:
            pe = self.proc_to_pe[proc]
            pe.writer = SimpleWriter(self, pe,
                                     self.output_mappings[proc],
                                     self.result_mappings)
            pe._write = types.MethodType(_simple_write, pe)
            provided_inputs = get_inputs(pe, inputs)
            try:
                other_inputs = all_inputs[proc]
                try:
                    provided_inputs.append(other_inputs)
                except:
                    provided_inputs = other_inputs
            except:
                pass

            if isinstance(provided_inputs, int):
                for i in range(0, provided_inputs):
                    _process_data(pe, {})
            else:

                if provided_inputs is None:
                    if not pe.inputconnections:
                        # run at least once for a source of the graph
                        provided_inputs = [{}]
                    else:
                        # no data
                        provided_inputs = []

                for data in provided_inputs:
                    _process_data(pe, data)

            for p, input_data in pe.writer.all_inputs.items():
                try:
                    all_inputs[p].extend(input_data)
                except:
                    all_inputs[p] = input_data
            if pe.writer.results:
                results[pe.id] = pe.writer.results
            # discard data from the PE writer
            pe.writer.all_inputs = {}
            pe.writer.results = {}
        results = self.map_outputs(results)
        return results


def _process_data(pe, data):
    # pe.log('Processing input: %s' % data)
    result = pe.process(data)
    # pe.log('Produced result: %s' % result)
    if result is not None:
        for output_name in result:
            pe.write(output_name, result[output_name])


def _simple_write(self, name, data):
    self.writer.write(name, data)


class SimpleWriter(object):
    def __init__(self, simple_pe, pe, output_mappings, result_mappings=None):
        self.simple_pe = simple_pe
        self.pe = pe
        self.output_mappings = output_mappings
        self.result_mappings = result_mappings
        self.all_inputs = {}
        self.results = {}

    def write(self, output_name, data):
        # self.pe.log('Writing %s to %s' % (data, output_name))
        try:
            destinations = self.output_mappings[output_name]
            dest_data = data
            for input_name, comm in destinations:
                copy_data = len(destinations) > 1 or len(comm.destinations) > 1
                for p in comm.destinations:
                    if copy_data:
                        dest_data = copy.deepcopy(dest_data)
                    input_data = {input_name: dest_data}
                    try:
                        self.all_inputs[p].append(input_data)
                    except:
                        self.all_inputs[p] = [input_data]
        except KeyError:
            # no destinations for this output
            # if there are no named result outputs
            # the data is added to the results of the PE
            if self.result_mappings is None:
                self.simple_pe.wrapper._write((self.pe.id, output_name),
                                              [data])
        # now check if the output is in the named results
        # (in case of a Tee) then data gets written to the PE results as well
        try:
            if output_name in self.result_mappings[self.pe.id]:
                self.simple_pe.wrapper._write((self.pe.id, output_name),
                                              [data])
        except:
            pass


def create_arg_parser():  # pragma: no cover
    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph for processing.')
    parser.add_argument('target', help='target execution platform')
    parser.add_argument('module', help='module that creates a dispel4py graph '
                        '(python module or file name)')
    parser.add_argument('-a', '--attr', metavar='attribute',
                        help='name of graph variable in the module')
    parser.add_argument('-f', '--file', metavar='inputfile',
                        help='file containing input dataset in JSON format')
    parser.add_argument('-d', '--data', metavar='inputdata',
                        help='input dataset in JSON format')
    parser.add_argument('-i', '--iter', metavar='iterations', type=int,
                        help='number of iterations', default=1)
    return parser


def create_inputs(args, graph):
    import json
    inputs = {}

    if args.file:
        if not os.path.exists(args.file):
            raise ValueError("File '%s' does not exist." % args.file)
        try:
            with open(args.file) as inputfile:
                inputs = json.loads(inputfile.read())
        except Exception as e:
            print("Error reading JSON file '%s': %s" % (args.file, str(e)))
            raise e
    elif args.data:
        inputs = json.loads(args.data)
    else:
        if args.iter == 1:
            print('Processing 1 iteration.')
        else:
            print('Processing %s iterations.' % args.iter)
        for node in graph.graph.nodes():
            if _is_root(node, graph):
                inputs[node.getContainedObject()] = args.iter

    # map input names to ids if necessary
    for node in graph.graph.nodes():
        pe = node.getContainedObject()
        try:
            d = inputs.pop(pe)
            inputs[pe.id] = d
        except:
            pass
        try:
            d = inputs.pop(pe.name)
            inputs[pe.id] = d
        except:
            pass

    return inputs


def load_graph_and_inputs(args):
    from dispel4py.utils import load_graph

    graph = load_graph(args.module, args.attr)
    if graph is None:
        return None, None

    graph.flatten()
    inputs = create_inputs(args, graph)
    return graph, inputs


def parse_common_args():   # pragma: no cover
    parser = create_arg_parser()
    return parser.parse_known_args()


def main():   # pragma: no cover
    from importlib import import_module

    args, remaining = parse_common_args()
    graph, inputs = load_graph_and_inputs(args)
    if graph is None:
        return

    try:
        # see if platform is in the mappings file as a simple name
        target = config[args.target]
    except KeyError:
        # it is a proper module name - fingers crossed...
        target = args.target
    try:
        parse_args = getattr(import_module(target), 'parse_args')
        args = parse_args(remaining, args)
    except SystemExit:
        # the sub parser raised an error
        raise
    except:
        # no other arguments required for target
        pass
    process = getattr(import_module(target), 'process')
    error_message = process(graph, inputs=inputs, args=args)
    if error_message:
        print(error_message)

if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a pipeline workflow with one producer
node and a chain of functions that process the data.

Execution:

* Simple processing:

    Execute the sequential mapping as follows::

        dispel4py simple dispel4py.examples.graph_testing.pipeline_composite\\
            [-i number of iterations]

    By default, if the number of iterations is not specified, the graph is
    executed once.

    For example::

        dispel4py simple dispel4py.examples.graph_testing.pipeline_composite

    Output::

        Processing 1 iteration.
        Inputs: {'TestProducer4': 1}
        SimplePE: Processed 1 iteration.
        Outputs: {'PE_subtract3': {'output': [5]}}


* MPI:

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 5 dispel4py mpi \\
            dispel4py.examples.graph_testing.pipeline_composite

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 5 nodes (4 function PEs and one producer). For this
        reason we need at least 5 MPI processes to execute it.

    Output::

        Processing 1 iteration.
        Processes: {'PE_addTwo0': [2], 'TestProducer4': [0], \
'PE_subtract3': [4], 'PE_multiplyByFour1': [1], 'PE_divideByTwo2': [3]}
        TestProducer4 (rank 0): Processed 1 iteration.
        PE_addTwo0 (rank 2): Processed 1 iteration.
        PE_multiplyByFour1 (rank 1): Processed 1 iteration.
        PE_divideByTwo2 (rank 3): Processed 1 iteration.
        PE_subtract3 (rank 4): Processed 1 iteration.

* STORM:

    From the dispel4py directory launch the Storm submission client::

        dispel4py storm dispel4py.examples.graph_testing.pipeline_composite\\
            -m remote
'''


from dispel4py.base import create_iterative_chain
from dispel4py.examples.graph_testing.testing_PEs import TestProducer
from dispel4py.workflow_graph import WorkflowGraph


def addTwo(data):
    '''
    Returns 2 + `data`.
    '''
    return 2 + data


def multiplyByFour(data):
    '''
    Returns 4 * `data`.
    '''
    return 4 * data


def divideByTwo(data):
    '''
    Returns `data`/2.
    '''
    return data/2


def subtract(data, n):
    '''
    Returns `data` - `n`.
    '''
    return data - n


functions = [addTwo, multiplyByFour, divideByTwo, (subtract, {'n': 1})]
composite = create_iterative_chain(functions)
producer = TestProducer()

graph = WorkflowGraph()
graph.connect(producer, 'output', composite, 'input')



'''
The mapping to Apache Spark.

Run as follows::

    spark-submit --py-files=<path to dispel4py binary distribution egg>
        dispel4py/new/spark_process.py <graph module or full path>
        [-f <input file>]
        [-d <input data>]
        [-a <graph attribute>]

The graph module must either be included in the python package with
`--py-files` and specified as a module name, or, if providing a filename,
the filename must be absolute and the file must be available on each node of
the Spark cluster.

For example::

    spark-submit --py-files=/path/to/dispel4py-1.0.1-py2.7.egg
        dispel4py/new/spark_process.py
        dispel4py.examples.graph_testing.pipeline_test -i 10

If the input JSON data contains a URI to a text file this will be read as a
Spark textfile input with one record per line.
'''

import argparse
import sys
import types


if sys.version_info == (3,):
    xrange = range


def simpleLogger(self, msg):
    print("%s: %s" % (self.id, msg))


class PEWrapper(object):

    def __init__(self, pe):
        self.pe = pe
        self.pe.log = types.MethodType(simpleLogger, pe)
        self.pe.preprocess()

    def process(self, data):
        # self.pe.log('processing %s' % data)
        for output, desc in self.pe.outputconnections.items():
            desc['writer'] = SimpleWriter(output)
        result = self.pe.process(data)
        # self.pe.log('result: %s' % result)
        written = []
        if result is not None:
            written.append(result)
        for output, desc in self.pe.outputconnections.items():
            written.extend(desc['writer'].data)
        # self.pe.log('writing: %s' % written)
        return written


class SimpleWriter(object):

    def __init__(self, output_name):
        self.output_name = output_name
        self.data = []

    def write(self, data):
        self.data.append({'output': data})


def parse_args():
    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph to Apache Spark.')
    parser.add_argument('module', help='module that creates a dispel4py graph '
                        '(python module or file name)')
    parser.add_argument('-a', '--attr', metavar='attribute',
                        help='name of graph variable in the module')
    parser.add_argument('-f', '--file', metavar='inputfile',
                        help='file containing input dataset in JSON format')
    parser.add_argument('-d', '--data', metavar='inputdata',
                        help='input dataset in JSON format')
    parser.add_argument('-i', '--iter', metavar='iterations', type=int,
                        help='number of iterations', default=1)
    parser.add_argument('-sm', '--master', help='master URL for the cluster')
    parser.add_argument('-sn', '--name', help='name of the Spark process')
    parser.add_argument(
        '-sd',
        '--deploy-mode',
        choices=['cluster', 'client'],
        help='deploy driver on worker nodes or locally as external client')
    result = parser.parse_args()
    return result


class Projection(object):

    def __init__(self, outputs):
        self.outputs = outputs

    def project(self, data):
        result = {}
        for o in self.outputs:
            if o in data:
                result[o] = data[o]
        if result:
            return [result]
        else:
            return []


class Rename(object):

    def __init__(self, mapping):
        self.mapping = mapping

    def rename(self, data):
        # print('renaming data: %s, mapping is: %s' % (data, self.mapping,))
        result = {}
        for o, i in self.mapping.items():
            if o in data:
                result[i] = data[o]
        if result:
            return [result]
        else:
            return []


def process(sc, workflow, inputs, args):

    from dispel4py.new.processor\
        import assign_and_connect, _order_by_dependency
    graph = workflow.graph
    result = assign_and_connect(workflow, graph.number_of_nodes())
    if result is None:
        return

    processes, inputmappings, outputmappings = result
    process_to_pes = {}
    wrappers = {}
    for node in workflow.graph.nodes():
        pe = node.getContainedObject()
        wrapper = PEWrapper(pe)
        for p in processes[pe.id]:
            process_to_pes[p] = pe
            wrappers[p] = wrapper
    # print('Processes: %s' % processes)
    # print(inputmappings)
    # print(outputmappings)
    ordered = _order_by_dependency(inputmappings, outputmappings)
    # print('Ordered processes: %s' % ordered)
    output_rdd = {}
    result_rdd = {}

    for proc in ordered:
        inps = inputmappings[proc]
        outs = outputmappings[proc]
        wrapper = wrappers[proc]
        pe = process_to_pes[proc]
        if inps:
            if len(inps) == 1:
                for input_name, sources in inps.items():
                    inp_rdd = output_rdd[(sources[0], input_name)]
                out_rdd = inp_rdd.flatMap(wrapper.process)
            else:
                prev = None
                for input_name, sources in inps.items():
                    inp_rdd = output_rdd[(sources[0], input_name)]
                    if prev:
                        inp_rdd = prev.union(inp_rdd)
                    prev = inp_rdd
                out_rdd = prev.flatMap(wrapper.process)
            if len(outs) == 1:
                for output_name in outs:
                    for inp in outs[output_name]:
                        input_name = inp[0]
                        rename = Rename({output_name: input_name})
                        output_rdd[(proc, input_name)] = \
                            out_rdd.flatMap(rename.rename)
            else:
                for output_name in outs:
                    proj = Projection([output_name])
                    proj_rdd = out_rdd.flatMap(proj.project)
                    for inp in outs[output_name]:
                        rename = Rename({output_name: inp[0]})
                        output_rdd[(proc, inp[0])] = \
                            proj_rdd.flatMap(rename.rename)
            if not outs:
                result_rdd[proc] = out_rdd

        else:
            pe_input = inputs[pe.id]
            if type(pe_input) is list:
                # only one slice so there no repetitions - not the best
                start_rdd = sc.parallelize(pe_input, 1)
            elif isinstance(pe_input, int):
                start_rdd = sc.parallelize(xrange(pe_input), 1)
            else:
                # fingers crossed it's a string and the file exists!
                start_rdd = sc.textFile(pe_input)
            out_rdd = start_rdd.flatMap(wrapper.process)
            if len(outs) == 1:
                for output_name in outs:
                    for inp in outs[output_name]:
                        input_name = inp[0]
                        rename = Rename({output_name: input_name})
                        output_rdd[(proc, input_name)] = \
                            out_rdd.flatMap(rename.rename)
            else:
                for output_name in outs:
                    proj = Projection([output_name])
                    out_rdd = out_rdd.flatMap(proj.project).persist()
                    for inp in outs[output_name]:
                        input_name = inp[0]
                        rename = Rename({output_name: input_name})
                        output_rdd[(proc, input_name)] = \
                            out_rdd.flatMap(rename.rename)
            if not outs:
                result_rdd[proc] = out_rdd
    # print("RESULT PROCESSES: %s" % result_rdd.keys())
    for p in result_rdd:
        result = result_rdd[p].foreach(lambda x: None)
        # result = result_rdd[p].foreach(lambda x:
        #                                simpleLogger(process_to_pes[p], x))
        # print 'RESULT FROM %s: %s' % (p, result)


def run():
    from pyspark import SparkContext, SparkConf

    conf = SparkConf()
    conf.setAppName('dispel4py')
    conf.set("spark.storage.memoryFraction", "0.5")
    sc = SparkContext(
        conf=conf)

    from dispel4py.new import processor
    from dispel4py.utils import load_graph

    args = parse_args()

    graph = load_graph(args.module, args.attr)
    if graph is None:
        return
    graph.flatten()

    inputs = processor.create_inputs(args, graph)

    process(sc, graph, inputs=inputs, args=args)


def main():
    import os
    import subprocess
    import sys
    try:
        spark_home = os.environ['SPARK_HOME']
    except:
        print('Please set SPARK_HOME.')
        sys.exit(1)

    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph for processing.')
    parser.add_argument('target', help='target execution platform')
    args, remaining = parser.parse_known_args()
    this_path = os.path.abspath(__file__)
    if this_path.endswith('pyc'):
        this_path = this_path[:-1]
    command = ['%s/bin/spark-submit' % spark_home,
               '--py-files=dist/dispel4py-1.0.1-py2.7.egg',
               this_path] + remaining
    print(command)
    subprocess.call(command)

if __name__ == '__main__':
    run()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Creates a Storm topology from a dispel4py graph.
'''

import json
import pickle

from storm import ttypes as tt

from dispel4py.workflow_graph import WorkflowNode
from dispel4py.core import GROUPING
from dispel4py.storm.utils import stormLogger

PE_WRAPPER = {'source': 'source_wrapper.py',
              'simple': 'simple_wrapper.py',
              'lockstep': 'lockstep_wrapper.py'}


def buildTopology(workflow):
    '''
    Builds a Storm topology from the given dispel4py workflow.
    '''
    workflow.propagate_types()
    graph = workflow.graph
    input_connections = {}
    input_mappings = {}
    node_names = {}

    node_counter = 1
    # give each node a unique name
    for node in graph.nodes():
        pe = node.getContainedObject()
        pe.log = stormLogger
        node_names[pe] = str(pe.__class__.__name__) + str(node_counter)
        node_counter += 1

    # now configure the connections
    for edge in graph.edges(data=True):
        direction = edge[2]['DIRECTION']
        source = direction[0]
        source_name = node_names[source]
        source_output = edge[2]['FROM_CONNECTION']
        dest = direction[1]
        dest_input = edge[2]['TO_CONNECTION']
        if dest not in input_connections:
            input_connections[dest] = {}
        if dest not in input_mappings:
            input_mappings[dest] = {}
        if source_name not in input_mappings[dest]:
            input_mappings[dest][source_name] = {}

        # find out the grouping of the input
        # shuffle grouping by default
        # this can be overridden by the PE implementation
        # or when creating the topology
        grouping = tt.Grouping(shuffle=tt.NullStruct())
        if GROUPING in dest.inputconnections[dest_input]:
            groupingtype = dest.inputconnections[dest_input][GROUPING]
            if isinstance(groupingtype, list):
                fields = []
                try:
                    # try if the grouping type is a list of indexes
                    for i in groupingtype:
                        fields.append(
                            source.getOutputTypes()[source_output][i])
                except TypeError:
                    # hopefully the grouping type is a list of
                    # tuple element names
                    fields = groupingtype
                # fields grouping with the list of fields
                grouping = tt.Grouping(fields=fields)
            elif groupingtype == 'all':
                grouping = tt.Grouping(all=tt.NullStruct())
            elif groupingtype == 'none':
                grouping = tt.Grouping(none=tt.NullStruct())
        input_connections[dest][
            tt.GlobalStreamId(source_name, source_output)] = grouping
        input_mappings[dest][source_name][source_output] = dest_input

    # add bolts and spouts
    spout_specs = {}
    bolt_specs = {}
    for node, data in graph.nodes(data=True):
        pe = node.getContainedObject()
        pe_name = node_names[pe]
        print("Spec'ing %s" % pe_name)
        # print pe
        # print pe.inputconnections
        # print pe.outputconnections
        # we're handling only PEs for now
        if node.nodeType != WorkflowNode.WORKFLOW_NODE_PE:
            raise Exception(
                "Unexpected workflow node of type '%s'" % node.nodeType)
        streams = {}
        for output, outtype in pe.getOutputTypes().items():
            streams[output] = tt.StreamInfo(outtype, False)

        # name of the file created in temporary directory resources
        module_name = pe.__module__
        pe_config = {}
        for key, value in vars(pe).items():
            if key not in pe.pickleIgnore:
                pe_config[key] = value

        json_conf = {'dispel4py.module': module_name,
                     'dispel4py.script': pe.__class__.__name__,
                     'dispel4py.inputmapping': {},
                     'dispel4py.config': pickle.dumps(pe_config)}
        if not pe.inputconnections or pe not in input_connections:
            # if there are no inputs it's a spout :P
            spout_specs[pe_name] = tt.SpoutSpec(
                spout_object=tt.ComponentObject(
                    shell=tt.ShellComponent("python", PE_WRAPPER['source'])),
                common=tt.ComponentCommon(
                    inputs={},
                    streams=streams,
                    json_conf=json.dumps(json_conf)))
        else:
            inputs = input_connections[pe] if pe in input_connections else {}
            json_conf['dispel4py.inputmapping'] = input_mappings[pe]\
                if pe in input_mappings else {}
            bolt_specs[pe_name] = tt.Bolt(
                bolt_object=tt.ComponentObject(
                    shell=tt.ShellComponent("python", PE_WRAPPER[pe.wrapper])),
                common=tt.ComponentCommon(
                    inputs=inputs,
                    streams=streams,
                    parallelism_hint=3,
                    json_conf=json.dumps(json_conf)))

    print("spouts %s" % spout_specs)
    print("bolts  %s" % bolt_specs)

    return tt.StormTopology(spouts=spout_specs,
                            bolts=bolt_specs,
                            state_spouts={})



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Tests for aggregation processing elements.

Using nose (https://nose.readthedocs.org/en/latest/) run as follows::

    nosetests dispel4py/new/aggregate_test.py
    ....
    ----------------------------------------------------------------------
    Ran 4 tests in 0.023s

    OK
'''

from dispel4py.examples.graph_testing.testing_PEs \
    import NumberProducer
from dispel4py.new.aggregate \
    import parallelSum, parallelCount, parallelMin, parallelMax, parallelAvg, \
    parallelStdDev, ContinuousReducePE

from dispel4py.new import simple_process
from dispel4py.workflow_graph import WorkflowGraph

from nose import tools


def graph_sum():
    prod = NumberProducer(1000)
    prod.name = 'NumberProducer'
    s = parallelSum()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', s, 'input')
    return graph


def graph_avg():
    prod = NumberProducer(1000)
    a = parallelAvg()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', a, 'input')
    return graph


def graph_min_max():
    prod = NumberProducer(1000)
    mi = parallelMin()
    ma = parallelMax()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', mi, 'input')
    graph.connect(prod, 'output', ma, 'input')
    return graph


def graph_count():
    prod = NumberProducer(1000)
    c = parallelCount()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', c, 'input')
    return graph


def graph_stddev():
    prod = NumberProducer(1000)
    std = parallelStdDev()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', std, 'input')
    return graph


def testSum():
    sum_wf = graph_sum()
    sum_wf.flatten()
    results = simple_process.process_and_return(
        sum_wf,
        inputs={'NumberProducer': [{}]})
    # there should be only one result
    tools.eq_(1, len(results))
    for key in results:
        tools.eq_({'output': [[499500]]}, results[key])


def testAvg():
    avg_wf = graph_avg()
    avg_wf.flatten()
    results = simple_process.process_and_return(
        avg_wf,
        inputs={'NumberProducer': [{}]})
    tools.eq_(1, len(results))
    for key in results:
        tools.eq_({'output': [[499.5, 1000, 499500]]}, results[key])


def testStdDev():
    stddev_wf = graph_stddev()
    stddev_wf.flatten()
    results = simple_process.process_and_return(
        stddev_wf,
        inputs={'NumberProducer': [{}]})
    tools.eq_(1, len(results))
    for key in results:
        # (41.51433802981722, 288.8182819698227, 1000, 499500)
        tools.eq_(1000, results[key]['output'][0][2])
        tools.eq_(499500, results[key]['output'][0][3])


def testCount():
    count_wf = graph_count()
    count_wf.flatten()
    results = simple_process.process_and_return(
        count_wf,
        inputs={'NumberProducer': [{}]})
    tools.eq_(1, len(results))
    for key in results:
        tools.eq_({'output': [[1000]]}, results[key])


def testMinMax():
    min_max_wf = graph_min_max()
    min_max_wf.flatten()
    results = simple_process.process_and_return(
        min_max_wf,
        inputs={'NumberProducer': [{}]})
    tools.eq_(2, len(results))
    for key in results:
        if key.startswith('MinPE'):
            tools.eq_({'output': [[0]]}, results[key])
        else:
            tools.eq_({'output': [[999]]}, results[key])


class TestPE(ContinuousReducePE):

    def __init__(self, indexes=[0]):
        ContinuousReducePE.__init__(self, indexes)

    def _process(self, data):
        for i in self.indexes:
            self.value[i] += data[i]


def testContinuousReduce():
    prod = NumberProducer()
    test = TestPE()
    graph = WorkflowGraph()
    graph.connect(prod, 'output', test, 'input')
    results = simple_process.process_and_return(graph, {prod: 5})
    tools.eq_({test.id: {'output': [[0] for i in range(5)]}}, results)


sum_wf = graph_sum()
avg_wf = graph_avg()
min_wf = graph_min_max()
count_wf = graph_count()
stddev_wf = graph_stddev()



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dispel4py.core import GenericPE
from dispel4py.base import IterativePE


class MyFirstPE(GenericPE):

    def __init__(self):
        GenericPE.__init__(self)
        self._add_input('input')
        self._add_output('prime')
        self._add_output('output')
        self.divisor = None

    def _process(self, inputs):
        number = inputs['input']
        if not self.divisor:
            self.divisor = number
            return {'prime': number}
        if not number % self.divisor == 0:
            return {'output': number}


from dispel4py.base import ProducerPE


class NumberProducer(ProducerPE):

    def __init__(self, limit):
        ProducerPE.__init__(self)
        self.limit = limit

    def _process(self):
        for i in range(2, self.limit):
            self.write(ProducerPE.OUTPUT_NAME, i)


class PrimeCollector(IterativePE):

    def __init__(self):
        IterativePE.__init__(self)

    def _process(self, data):
        return data


from dispel4py.workflow_graph import WorkflowGraph


graph = WorkflowGraph()
producer = NumberProducer(1000)
primes = PrimeCollector()
prev = producer
for i in range(2, 200):
    divide = MyFirstPE()
    graph.connect(prev, 'output', divide, 'input')
    prev = divide
    graph.connect(divide, 'prime', primes, 'input')



from dispel4py.examples.graph_testing.testing_PEs\
    import TestMultiProducer, TestOneInOneOut
from dispel4py.workflow_graph import WorkflowGraph


prod = TestMultiProducer()
cons1 = TestOneInOneOut()
cons2 = TestOneInOneOut()

graph = WorkflowGraph()
graph.connect(prod, 'output', cons1, 'input')
graph.connect(cons1, 'output', cons2, 'input')



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Enactment of dispel4py graphs using multiprocessing.

From the commandline, run the following command::

    dispel4py multi <module> -n num_processes [-h] [-a attribute]\
                    [-f inputfile] [-i iterations]

with parameters

:module:    module that creates a Dispel4Py graph
:-n num:    number of processes (required)
:-a attr:   name of the graph attribute within the module (optional)
:-f file:   file containing input data in JSON format (optional)
:-i iter:   number of iterations to compute (default is 1)
:-h:        print this help page

For example::

    dispel4py multi dispel4py.examples.graph_testing.pipeline_test -i 5 -n 6
    Processing 5 iterations.
    Processes: {'TestProducer0': [5], 'TestOneInOneOut5': [2],\
                'TestOneInOneOut4': [4], 'TestOneInOneOut3': [3],\
                'TestOneInOneOut2': [1], 'TestOneInOneOut1': [0]}
    TestProducer0 (rank 5): Processed 5 iterations.
    TestOneInOneOut1 (rank 0): Processed 5 iterations.
    TestOneInOneOut2 (rank 1): Processed 5 iterations.
    TestOneInOneOut3 (rank 3): Processed 5 iterations.
    TestOneInOneOut4 (rank 4): Processed 5 iterations.
    TestOneInOneOut5 (rank 2): Processed 5 iterations.
'''


import argparse
import copy
import multiprocessing
import traceback
import types
from dispel4py.new.processor \
    import GenericWrapper, simpleLogger, STATUS_ACTIVE, STATUS_TERMINATED
from dispel4py.new import processor


def _processWorker(wrapper):
    wrapper.process()


def parse_args(args, namespace):    # pragma: no cover
    parser = argparse.ArgumentParser(
        prog='dispel4py',
        description='Submit a dispel4py graph to multiprocessing.')
    parser.add_argument('-s', '--simple', help='force simple processing',
                        action='store_true')
    parser.add_argument('-n', '--num', metavar='num_processes', required=True,
                        type=int, help='number of processes to run')
    result = parser.parse_args(args, namespace)
    return result


def process(workflow, inputs, args):
    size = args.num
    success = True
    nodes = [node.getContainedObject() for node in workflow.graph.nodes()]
    if not args.simple:
        try:
            result = processor.assign_and_connect(workflow, size)
            processes, inputmappings, outputmappings = result
        except:
            success = False

    if args.simple or not success:
        ubergraph = processor.create_partitioned(workflow)
        print('Partitions: %s' % ', '.join(('[%s]' % ', '.join(
            (pe.id for pe in part)) for part in workflow.partitions)))
        for node in ubergraph.graph.nodes():
            wrapperPE = node.getContainedObject()
            pes = [n.getContainedObject().id for
                   n in wrapperPE.workflow.graph.nodes()]
            print('%s contains %s' % (wrapperPE.id, pes))

        try:
            result = processor.assign_and_connect(ubergraph, size)
            if result is None:
                return 'dispel4py.multi_process: ' \
                       'Not enough processes for execution of graph'
            processes, inputmappings, outputmappings = result
            inputs = processor.map_inputs_to_partitions(ubergraph, inputs)
            success = True
            nodes = [node.getContainedObject()
                     for node in ubergraph.graph.nodes()]
        except:
            print(traceback.format_exc())
            return 'dispel4py.multi_process: ' \
                   'Could not create mapping for execution of graph'

    print('Processes: %s' % processes)

    process_pes = {}
    queues = {}
    result_queue = None
    try:
        if args.results:
            result_queue = multiprocessing.Queue()
    except AttributeError:
        pass
    for pe in nodes:
        provided_inputs = processor.get_inputs(pe, inputs)
        for proc in processes[pe.id]:
            cp = copy.deepcopy(pe)
            cp.rank = proc
            cp.log = types.MethodType(simpleLogger, cp)
            wrapper = MultiProcessingWrapper(proc, cp, provided_inputs)
            process_pes[proc] = wrapper
            wrapper.input_queue = multiprocessing.Queue()
            wrapper.input_queue.name = 'Queue_%s_%s' % (cp.id, cp.rank)
            wrapper.result_queue = result_queue
            queues[proc] = wrapper.input_queue
            wrapper.targets = outputmappings[proc]
            wrapper.sources = inputmappings[proc]
    for proc in process_pes:
        wrapper = process_pes[proc]
        wrapper.output_queues = {}
        for target in wrapper.targets.values():
            for inp, comm in target:
                for i in comm.destinations:
                    wrapper.output_queues[i] = queues[i]

    jobs = []
    for wrapper in process_pes.values():
        p = multiprocessing.Process(target=_processWorker, args=(wrapper,))
        jobs.append(p)

    for j in jobs:
        j.start()

    for j in jobs:
        j.join()

    if result_queue:
        result_queue.put(STATUS_TERMINATED)
    return result_queue


class MultiProcessingWrapper(GenericWrapper):

    def __init__(self, rank, pe, provided_inputs=None):
        GenericWrapper.__init__(self, pe)
        self.pe.log = types.MethodType(simpleLogger, pe)
        self.pe.rank = rank
        self.provided_inputs = provided_inputs
        self.terminated = 0

    def _read(self):
        result = super(MultiProcessingWrapper, self)._read()
        if result is not None:
            return result
        # read from input queue
        no_data = True
        while no_data:
            try:
                data, status = self.input_queue.get()
                no_data = False
            except:
                self.pe.log('Failed to read item from queue')
                pass
        while status == STATUS_TERMINATED:
            self.terminated += 1
            if self.terminated >= self._num_sources:
                return data, status
            else:
                try:
                    data, status = self.input_queue.get()
                except:
                    self.pe.log('Failed to read item from queue')
                    pass
        return data, status

    def _write(self, name, data):
        # self.pe.log('Writing %s to %s' % (data, name))
        try:
            targets = self.targets[name]
        except KeyError:
            # no targets
            if self.result_queue:
                self.result_queue.put((self.pe.id, name, data))
            return
        for (inputName, communication) in targets:
            output = {inputName: data}
            dest = communication.getDestination(output)
            for i in dest:
                # self.pe.log('Writing out %s' % output)
                try:
                    self.output_queues[i].put((output, STATUS_ACTIVE))
                except:
                    self.pe.log("Failed to write item to output '%s'" % name)

    def _terminate(self):
        for output, targets in self.targets.items():
            for (inputName, communication) in targets:
                for i in communication.destinations:
                    self.output_queues[i].put((None, STATUS_TERMINATED))


def main():    # pragma: no cover
    from dispel4py.new.processor \
        import load_graph_and_inputs, parse_common_args

    args, remaining = parse_common_args()
    args = parse_args(remaining, args)

    graph, inputs = load_graph_and_inputs(args)
    if graph is not None:
        errormsg = process(graph, inputs, args)
        if errormsg:
            print(errormsg)



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Enactment of dispel4py graphs with MPI.

From the commandline, run the following command::

    dispel4py mpi <module> [-h] [-a attribute] [-f inputfile] [-i iterations]

with parameters

:module:    module that creates a Dispel4Py graph
:-a attr:   name of the graph attribute within the module (optional)
:-f file:   file containing input data in JSON format (optional)
:-i iter:   number of iterations to compute (default is 1)
:-h:        print this help page

For example::

    mpiexec -n 6 dispel4py mpi dispel4py.examples.graph_testing.pipeline_test\
        -i 5
    Processing 5 iterations.
    Processing 5 iterations.
    Processing 5 iterations.
    Processing 5 iterations.
    Processing 5 iterations.
    Processing 5 iterations.
    Processes: {'TestProducer0': [5], 'TestOneInOneOut5': [2],\
        'TestOneInOneOut4': [4], 'TestOneInOneOut3': [3],\
        'TestOneInOneOut2': [1], 'TestOneInOneOut1': [0]}
    TestOneInOneOut1 (rank 0): Processed 5 iterations.
    TestOneInOneOut2 (rank 1): Processed 5 iterations.
    TestOneInOneOut3 (rank 3): Processed 5 iterations.
    TestProducer0 (rank 5): Processed 5 iterations.
    TestOneInOneOut4 (rank 4): Processed 5 iterations.
    TestOneInOneOut5 (rank 2): Processed 5 iterations.
'''

from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

from dispel4py.new.processor\
    import GenericWrapper, simpleLogger, STATUS_TERMINATED, STATUS_ACTIVE
from dispel4py.new import processor

import argparse
import sys
import types
import traceback


def mpi_excepthook(type, value, trace):
    '''
    Sending abort to all processes if an exception is raised.
    '''
    if rank == 0:
        traceback.print_tb(trace)
    comm.Abort(1)


sys.excepthook = mpi_excepthook


def parse_args(args, namespace):
    parser = argparse.ArgumentParser(
        description='Submit a dispel4py graph to MPI processes.')
    parser.add_argument('-s', '--simple', help='force simple processing',
                        action='store_true')
    result = parser.parse_args(args, namespace)
    return result


def process(workflow, inputs, args):
    processes = {}
    inputmappings = {}
    outputmappings = {}
    success = True
    nodes = [node.getContainedObject() for node in workflow.graph.nodes()]
    if rank == 0 and not args.simple:
        try:
            processes, inputmappings, outputmappings =\
                processor.assign_and_connect(workflow, size)
        except:
            success = False
    success = comm.bcast(success, root=0)

    if args.simple or not success:
        ubergraph = processor.create_partitioned(workflow)
        nodes = [node.getContainedObject() for node in ubergraph.graph.nodes()]
        if rank == 0:
            print('Partitions: %s' % ', '.join(('[%s]' % ', '.join(
                (pe.id for pe in part)) for part in workflow.partitions)))
            for node in ubergraph.graph.nodes():
                wrapperPE = node.getContainedObject()
                print('%s contains %s' % (wrapperPE.id,
                                          [n.getContainedObject().id for n in
                                           wrapperPE.workflow.graph.nodes()]))
            try:
                processes, inputmappings, outputmappings =\
                    processor.assign_and_connect(ubergraph, size)
                inputs = processor.map_inputs_to_partitions(ubergraph, inputs)
                success = True
            except:
                # print traceback.format_exc()
                print('dispel4py.mpi_process: \
                    Not enough processes for execution of graph')
                success = False

    success = comm.bcast(success, root=0)

    if not success:
        return

    try:
        inputs = {pe.id: v for pe, v in inputs.items()}
    except AttributeError:
        pass
    processes = comm.bcast(processes, root=0)
    inputmappings = comm.bcast(inputmappings, root=0)
    outputmappings = comm.bcast(outputmappings, root=0)
    inputs = comm.bcast(inputs, root=0)

    if rank == 0:
        print('Processes: %s' % processes)
        # print 'Inputs: %s' % inputs

    for pe in nodes:
        if rank in processes[pe.id]:
            provided_inputs = processor.get_inputs(pe, inputs)
            wrapper = MPIWrapper(pe, provided_inputs)
            wrapper.targets = outputmappings[rank]
            wrapper.sources = inputmappings[rank]
            wrapper.process()


class MPIWrapper(GenericWrapper):

    def __init__(self, pe, provided_inputs=None):
        GenericWrapper.__init__(self, pe)
        self.pe.log = types.MethodType(simpleLogger, pe)
        self.pe.rank = rank
        self.provided_inputs = provided_inputs
        self.terminated = 0

    def _read(self):
        result = super(MPIWrapper, self)._read()
        if result is not None:
            return result

        status = MPI.Status()
        msg = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        tag = status.Get_tag()
        while tag == STATUS_TERMINATED:
            self.terminated += 1
            if self.terminated >= self._num_sources:
                break
            else:
                msg = comm.recv(source=MPI.ANY_SOURCE,
                                tag=MPI.ANY_TAG,
                                status=status)
                tag = status.Get_tag()
        return msg, tag

    def _write(self, name, data):
        try:
            targets = self.targets[name]
        except KeyError:
            # no targets
            # self.pe.log('Produced output: %s' % {name: data})
            return
        for (inputName, communication) in targets:
            output = {inputName: data}
            dest = communication.getDestination(output)
            for i in dest:
                try:
                    # self.pe.log('Sending %s to %s' % (output, i))
                    request = comm.isend(output, tag=STATUS_ACTIVE, dest=i)
                    status = MPI.Status()
                    request.Wait(status)
                except:
                    self.pe.log(
                        'Failed to send data stream "%s" to rank %s: %s'
                        % (name, i, traceback.format_exc()))

    def _terminate(self):
        for output, targets in self.targets.items():
            for (inputName, communication) in targets:
                for i in communication.destinations:
                    # self.pe.log('Terminating consumer %s' % i)
                    comm.isend(None, tag=STATUS_TERMINATED, dest=i)


def main():
    from dispel4py.new.processor \
        import load_graph_and_inputs, parse_common_args

    args, remaining = parse_common_args()
    try:
        args = parse_args(remaining, args)
    except SystemExit:
        raise

    graph, inputs = load_graph_and_inputs(args)
    if graph is not None:
        errormsg = process(graph, inputs, args)
        if errormsg:
            print(errormsg)



from dispel4py.workflow_graph import WorkflowGraph
from dispel4py.examples.graph_testing.testing_PEs \
    import TestProducer, TestOneInOneOut

graph = WorkflowGraph()
prod = TestProducer()
cons1 = TestOneInOneOut()
cons2 = TestOneInOneOut()
cons3 = TestOneInOneOut()
cons4 = TestOneInOneOut()

graph.connect(prod, 'output', cons1, 'input')
graph.connect(prod, 'output', cons2, 'input')
graph.connect(cons1, 'output', cons3, 'input')
graph.connect(cons3, 'output', cons4, 'input')
graph.connect(cons4, 'output', cons2, 'input')



# Copyright (c) The University of Edinburgh 2014-2015
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This is a dispel4py graph which produces a workflow that copies the data
(from node prod) to two nodes (cons1 and cons2).
If you compare this graph with :py:mod:`~test.graph_testing.grouping_onetoall`
they look quite similar. However, they do different things.
In this example, the nodes ``cons1`` and ``cons2`` are different PE and the
node ``prod`` sends the same data to both PEs.

.. image:: /images/teecopy.png

It can be executed with MPI and STORM.

* MPI: Please, locate yourself into the dispel4py directory.

    Execute the MPI mapping as follows::

        mpiexec -n <number mpi_processes> dispel4py mpi <module|module file>\\
            [-a name_dispel4py_graph]\\
            [-f file containing the input dataset in JSON format]\\
            [-i number of iterations/runs']\\
            [-s]

    The argument '-s' forces to run the graph in a simple processing, which
    means that the first node of the graph will be executed in a process, and
    the rest of nodes will be executed in a second process.
    When [-i number of interations/runs] is not indicated, the graph is
    executed once by default.


    For example::

        mpiexec -n 3 dispel4py mpi dispel4py.examples.graph_testing.teecopy

    .. note::

        Each node in the graph is executed as a separate MPI process.
        This graph has 3 nodes. For this reason we need at least 3 MPI
        processes to execute it.

    Output::

        Processing 1 iteration.
        Processes: {'TestProducer0': [0], 'TestOneInOneOut2': [2], \
'TestOneInOneOut1': [1]}
        TestProducer0 (rank 0): Processed 1 iteration.
        TestOneInOneOut2 (rank 2): Processed 1 iteration.
        TestOneInOneOut1 (rank 1): Processed 1 iteration.
'''

from dispel4py.examples.graph_testing import testing_PEs as t
from dispel4py.workflow_graph import WorkflowGraph


def testTee():
    '''
    Creates a graph with two consumer nodes and a tee connection.

    :rtype: the created graph
    '''
    graph = WorkflowGraph()
    prod = t.TestProducer()
    cons1 = t.TestOneInOneOut()
    cons2 = t.TestOneInOneOut()
    graph.connect(prod, 'output', cons1, 'input')
    graph.connect(prod, 'output', cons2, 'input')
    return graph

''' important: this is the graph_variable '''
graph = testTee()



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

sys.path.append('resources')

from dispel4py.storm.client import StormClient

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol
from storm import ttypes as tt


if __name__ == '__main__':

    with open("topology.thrift", "r") as thrift_file:
        bytes = thrift_file.read()

    transportIn = TTransport.TMemoryBuffer(bytes)
    protocolIn = TBinaryProtocol.TBinaryProtocol(transportIn)
    topology = tt.StormTopology()
    topology.read(protocolIn)

    topology_name, host, port, uploaded_jar_location = sys.argv[1:]

    conf = {'topology.workers': 10, 'topology.max.spout.pending': 20}
    client = StormClient(host, port)
    try:
        print("Submitting topology '%s' to %s:%s ... "
              % (topology_name, host, port))
        client.submitTopology(topology_name,
                              uploaded_jar_location,
                              conf,
                              topology)
    except tt.AlreadyAliveException:
        print("Submission failed: Topology '%s' already exists."
              % topology_name)



# Copyright (c) The University of Edinburgh 2014
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="dispel4py",
    version="1.2.1",
    author="The University of Edinburgh",
    author_email="a.krause@epcc.ed.ac.uk",
    description=("dispel4py is a free and open-source Python library for \
        describing abstract stream-based workflows for distributed \
        data-intensive applications."),
    license="Apache 2",
    keywords="dispel4py dispel workflows processing elements data intensive",
    url="http://dispel4py.org",
    packages=['dispel4py',
              'dispel4py.new',
              'dispel4py.storm',
              'dispel4py.examples',
              'dispel4py.examples.graph_testing'],
    long_description=read('README.md'),
    install_requires=['networkx', 'flake8'],
    entry_points={
        'console_scripts': [
            'dispel4py = dispel4py.new.processor:main',
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache 2 License",
    ],
)
